(* This library defines an executable version of the DFS algorithm,
   and proves that it is correct with respect to the predicate [dfs].
   This should be a simple matter, but justifying that the algorithm
   terminates requires a little dependent-type trickery. *)

Set Implicit Arguments.
Require Import Coq.Wellfounded.Inverse_Image.
Require Import List.
Require Import LibTactics.
Require Import MyTactics.
Require Import MyFix.
Require Import MySets.
Require Import MyRelations.
Require Import DFS.
Require Import SigSET.
Require Import SigFOREACH.

(* ---------------------------------------------------------------------------- *)

Section DFSExec.

  (* We assume that the graph is defined, at the mathematical level, by a set
     of vertices [V] and a set of edges [E]. We assume that [V] is finite. *)

  Variable V : Type.
  Variable E : V -> V -> Prop.

  Variable finiteness:
    finite (@universe V).

  (* We assume that [V] is also the type of machine representations of
     vertices. This simplifies our life, and does not seem to be a very
     restrictive assumption. *)

  (* We assume that sets of vertices are represented at runtime as dictated
     by the parameter [VSet]. *)

  Variable VSet : SET V.

  Notation repr := (repr VSet).
  Notation meaning := (meaning VSet).
  Notation void := (void VSet).
  Notation mark := (mark VSet).
  Notation marked := (marked VSet).

  (* We assume that a function [foreach successor] allows iterating over all
     successors of a vertex [v]. *)

  Variable successor : V -> FOREACH V.
  Variable successor_spec :
    forall v,
    FOREACH_SET_SPEC (successor v) (image E (singleton v)).

(* ---------------------------------------------------------------------------- *)

  (* The set of marked vertices grows with time. We have assumed that [V] is
     finite: hence, this set cannot grow forever. This implies that the DFS
     algorithm terminates. *)

  Definition le marks1 marks2 :=
    subset (meaning marks1) (meaning marks2).

  Notation gt marks2 marks1 :=
    (strict_subset (meaning marks1) (meaning marks2)).

  Lemma wf_gt:
    well_founded (fun x y => gt x y).
  Proof.
    eapply wf_inverse_image with (R := fun vs ws : set V => strict_subset ws vs).
    eapply wf_strict_superset. exact finiteness.
  Qed.

  (* If a vertex [v] is not marked, then marking it causes the set of marked
     vertices to grow strictly. *)

  Lemma prove_lt:
    forall v marks,
    marked v marks = false ->
    gt (mark v marks) marks.
  Proof.
    intros. rewrite mark_spec.
    eapply prove_strict_subset; eauto using decision_procedure_says_no, marked_spec with mysets.
  Qed.

  (* The orderings [gt] and [le], which bear on sets of marks, are sometimes
     lifted to states, where a state is a pair of a set of marks and a forest.
     The second component of the pair is irrelevant. *)

  Definition pi1 (A1 A2 : Type) (a : A1 * A2) : A1 :=
    let (a1, _) := a in a1.

  Definition lift (A1 A2 : Type) (R : A1 -> A1 -> Prop) (a b : A1 * A2) :=
    R (pi1 a) (pi1 b).

  Lemma wf_lift:
    forall (A1 A2 : Type) (R : A1 -> A1 -> Prop),
    well_founded R ->
    well_founded (@lift A1 A2 R).
  Proof.
    unfold lift. eauto using wf_inverse_image.
  Qed.

(* ---------------------------------------------------------------------------- *)

  (* We now define the recursive function [visitf_dep], which is the heart of
     the DFs algorithm. This algorithm maintains a current state which is a
     pair of a set of marked vertices and a forest. As the algorithm runs, the
     set of marked vertices grows, and a tree whose root is [v] is prepended
     to the forest, if [v] was not marked already. The call [visitf_dep s0 v]
     has the effect of visiting the vertex [v] and returning an updated state
     [s1]. *)

  Notation state :=
    (repr * forest V)%type.

  Ltac state :=
    repeat match goal with s: state |- _ => destruct s as [ ? ? ] end.

  (* The states [s0] and [s1] are related by [le]. We build this claim into
     the result type of [visitf_dep], as this is required (I believe) for the
     termination argument to go through. We define a non-dependent version,
     [visitf], later on. *)

  Definition visitf_dep:
    forall s0 : state,
    V ->
    { s1 | lift le s0 s1 }.
  Proof.

    (* The construction is by well-founded induction. The set of marked
       vertices grows strictly at each recursive invocation. *)
    eapply (Fix (wf_lift wf_gt) (fun s0 => V -> { s1 | lift le s0 s1 })).

    (* The initial state is [s0], also known as the pair [marks0, forest0].
       The function itself, for recursive calls, is [visitf_dep]. The vertex
       to be visited is [v]. *)
    (* We could simply destruct [s0] as [marks0, forest0]. We use [case_eq]
       instead because this allows us to avoid re-building this pair in the
       first branch below. *)
    intros s0 visitf_dep v. case_eq s0; intros marks0 forest0 h0.

    (* The code begins with a test: is [v] marked? *)
    case_eq (marked v marks0); intro heq.

    (* First branch. [v] is marked. There is nothing to do: we return the
       initial state [s0]. *)
    exists s0.
      (* Proof obligation for the first branch. We must prove that the
         state has evolved along [lift le], which is immediate, since no
         state change has taken place. For clarity, we make this sub-proof
         opaque, and are careful to avoid an unnecessary dependency on [heq]. *)
      clear heq. abstract (rewrite h0; eapply subset_reflexive).

    (* Second branch. [v] is not marked. We must first update the state by
       marking [v] and resetting the current forest to be [Empty]. Then, we
       examine the successors of [v], yielding updated marks and a forest of
       children [vs]. Finally, we must prepend the tree [v/vs] in front of the
       forest [forest0] that was passed to us. *)

    (* We now call [successors], then perform a post-treatment, which is
       partly relevant (construct a new forest node) and partly erased (there
       is a proof obligation, which involves transitivity). *)

    assert (s1: { s1 | lift le (mark v marks0, (Empty V)) s1 }).

    (* This is the call to [successors]. Note that the type [A] at which it is
       used is the dependent type [{ s1 | lift le (mark v marks0, _) s1 }].
       That is, we keep track of the fact that, throughout the loop, the
       current set of marks remains a superset of [mark v marks0], hence a
       strict superset of [marks0]. This allows us to justify the recursive
       call to [visitf_dep]. *)

    eapply (foreach (successor v)).
      (* The initial state argument is formed by pairing [marks v marks0]
         with an empty sub-forest. *)
      eexists. eapply subset_reflexive.
      (* We now define what must be done at each successor [w] of [v].
         This function is invoked in a state [s1], which is a pair of
         [marks1] and [forest1]. *)
      intros [ s1 hle1 ] w. case_eq s1; intros marks1 forest1 h1.
        (* We apply [visitf_dep] to the current state [s1] and to [w]. *)
        assert (s2 : {s2 | lift le s1 s2}).
          eapply (visitf_dep s1); [ | exact w ].
          (* We must check that there are more marked vertices in the
             current state [s1] than in the initial state [s0]. *)
          generalize (prove_lt _ _ heq). clear heq.
          abstract (unfold lift, le; intros; subst; eauto using strict_subset_transitive_1).
        (* The call [visitf_dep s1 w] produces a state [s2] which is known
           to be above [s1]. We conclude, by transitivity, that it is also
           above [(mark v marks0, _)]. *)
        destruct s2 as [ s2 ? ].
        exists s2. eapply subset_transitive; eassumption.

    (* This is the post-treatment. We have two independent things to do.
       The set of marked vertices is transmitted unchanged, but we must
       argue that, if it is above [marks v marks0], then it is also above
       [marks0]. The forest [vs] produced by [successors] represents the
       children of [v]. We build the tree [v/vs] and prepend it in front
       of the forest that was originally passed to us, [forest0]. *)

    destruct s1 as [ [ marks1 forest1 ] h1 ].
    exists (marks1, NonEmpty v forest1 forest0).
      (* Proof obligation. The application [prove_lt heq] uses the equality
         assumption [heq], so we need it to make it explicitly visible; we
         must not bury it inside this opaque sub-proof. *)
      generalize (prove_lt _ _ heq). clear heq.
      abstract (unfold lift, le; eauto using strict_subset_implies_subset with st).

  Defined.

(* ---------------------------------------------------------------------------- *)

  (* We now write down the specification of [visitf_dep] and [visitf]. It is a
     specification also for iterated applications of these functions, i.e.,
     for [fold_left visitf_dep] and [fold_left visitf]. *)

  Definition visitf_spec (s : state) (vs : set V) (s' : state) :=

    (* Assume that the function is invoked in the state [(marks, forest)] at a
       set of root vertices [vs] and yields the state [(marks', forest')]. *)

    let (marks, forest) := s in
    let (marks', forest') := s' in
      
    (* Then, we expect the following properties to hold: *)

    (* 0. More vertices are marked in the final state. *)
    le marks marks' /\

    (* 1. The vertices of [vs] are marked in the final state. *)
    subset vs (meaning marks') /\

    (* 2. The roots of [forest'] are the roots of [forest], plus possibly
       some of the members of [vs]. *)
    subset (roots forest') (union vs (roots forest)) /\

    (* 3. and, assuming that [forest] was a valid DFS forest with respect to
          the initial state, then [forest'] is a valid DFS forest with respect
          to the final state. For obvious efficiency reasons, our forests are
          built reversed: the last discovered tree appears at the head of the
          forest. *)
    forall init,
    dfs E init (meaning marks ) (revf forest ) ->
    dfs E init (meaning marks') (revf forest').

    (* Points 1. and 2. above are not of great interest, but they are used as
       part of the proof of point 3. Point 0 itself is used in the proof of
       point 1. *)

    (* Point 3. could be equivalently stated as follows: [revf forest'] is the
       concatenation of [revf forest] with some valid DFS forest with respect
       to the initial and final states [marks] and [marks']. The formulation
       above avoids the existential quantifier and seems more illuminating. *)

  (* [visitf_spec] is reflexive. That is, if the vertices [vs] are already
     marked, then no state change occurs. This includes the case where the
     set [vs] is empty. *)

  Lemma visitf_spec_reflexive:
    forall marks forest vs,
    subset vs (meaning marks) ->
    visitf_spec (marks, forest) vs (marks, forest).
  Proof.
    unfold visitf_spec, le. eauto with mysets.
  Qed.

  (* [visitf_spec] is transitive. In order to visit the roots [union vs ws]
     it suffices to visit successively the roots [vs] and [ws]. *)

  Lemma visitf_spec_transitive:
    forall s0 vs s1 ws s2,
    visitf_spec s0 vs s1 ->
    visitf_spec s1 ws s2 ->
    visitf_spec s0 (union vs ws) s2.
  Proof.
    unfold visitf_spec. intros. state. unpack. splits.
    (* 0. By transitivity of [le]. *)
    eapply subset_transitive; eassumption.
    (* 1. By the corresponding hypotheses and by monotonicity (point 0). *)
    solve [ eauto with st mysets ].
    (* 2. By the corresponding hypotheses. *)
    eapply subset_transitive; [ eassumption | ].
    eapply prove_subset_union_left; [ eauto with mysets | ].
    eapply subset_transitive; [ eassumption | ].
    solve [ eauto with mysets ].
    (* 3. Immediate. *)
    eauto.
  Qed.

  (* In order to visit a single unmarked root [v], it suffices to mark [v];
     visit its successors [ws], beginning with an empty forest and yielding
     a sub-forest [children]; and to finally construct a new tree by placing
     the vertex [v] above the sub-forest [children]. The forest [forest] is
     an arbitrary previously-constructed forest in front of which we are
     placing this new tree. *)

  Lemma visitf_spec_unmarked:
    forall marks v ws marks' children forest, 
    marked v marks = false ->
    outof E (singleton v) ws ->
    into E (singleton v) ws ->
    visitf_spec (mark v marks, Empty _) ws (marks', children) ->
    visitf_spec (marks, forest) (singleton v) (marks', NonEmpty v children forest).
  Proof.
    introv hv houtof hinto [ h0 [ h1 [ h2 h3 ]]]. simpl in h2, h3.
    splits.
    (* 0. The set of marked vertices grows. *)
    eapply subset_transitive; [ | eexact h0 ].
    rewrite mark_spec. solve [ eauto with mysets ].
    (* 1. [v] is marked before we visit its successors, so it remains marked. *)
    eapply subset_transitive.
      instantiate (1 := insert v (meaning marks)). solve [ eauto with mysets ].
      rewrite <- mark_spec. exact h0.
    (* 2. [v] becomes a new root of the new forest. *)
    solve [ eauto with mysets ].
    (* 3. Simplify the application of [revf] to this non-empty forest, then use
       the lemma [dfs_concat] to simplify the goal. Conclude by applying the
       constructor [DFSNonEmpty] and checking that its premises are satisfied. *)
    introv hdfs. simpl.
    eapply dfs_concat; [ eexact hdfs | clear forest init hdfs ].
    econstructor.
      eauto using decision_procedure_says_no, marked_spec.
      rewrite <- mark_spec. eauto with dfs.
      rewrite roots_revf. solve [ eauto with st mysets ].
      solve [ eauto with st mysets ].
      econstructor.
  Qed.

  (* The above three lemmas represent the meat of the correctness proof for
     [visitf] and [fold_left visitf]. The rest is just boilerplate, so to
     speak; or, more precisely, extraction of proof obligations out of the
     code. *)

(* ---------------------------------------------------------------------------- *)

  (* We now prove a key lemma which represents a step towards the proof of the
     correctness of [visitf_dep]. This lemma is parameterized over
     [visitf_dep], under the name [self]. It states that, if [self] satisfies
     the specification of [visitf_dep], then [fold_left self ws] satisfies the
     same specification. The statement of this lemma can be thought of as an
     invariant for the loop over the successors of [v] in the definition of
     [visitf_dep]. *)

  Section Self.

    (* The function [self] does not take a raw state, but a fat state, of type
       { s: state | P s }, for some property [P] which is irrelevant here. *)

    Variable P : state -> Prop.
    Notation fat_state := (sig P).
    Notation inject  := (exist P).

    Variable self : fat_state -> V -> fat_state.

    (* [self] must satisfy the following specification. *)
    Variable self_spec:
      forall s v s' pf pf',
      self (inject s pf) v = inject s' pf' ->
      visitf_spec s (singleton v) s'.

    (* Here comes the key lemma. *)

    Lemma fold_self_correct:
      forall s' ws s pf' pf,
      fold_left self ws (inject s pf) = inject s' pf' ->
      visitf_spec s (lsupport ws) s'.
    Proof.
      (* This is a straightforward statement about a loop. We prove it by
         structural induction over the list [ws]. Modulo the boilerplate,
         the proof reduces to the fact that [visitf_spec] is reflexive and
         transitive. *)
      induction ws as [| w ws ]; simpl; introv hfold.
      (* Case Nil. *)
      xx. state.
      solve [ eauto using visitf_spec_reflexive with mysets ].
      (* Case Cons. *)
      (* We are looking at a call to [self]. Name its final state. *)
      match type of hfold with context[self ?s0 w] =>
        destruct (self s0 w) as [ [ marks1 forest1 ] pf1 ] _eqn
      end.
      (* Next, we are looking at a recursive call to [fold]. Name its final state. *)
      match type of hfold with context[fold_left self ws ?s1] =>
        destruct (fold_left self ws s1) as [ [ marks2 forest2 ] pf2 ] _eqn
      end.
      xx.
      (* By the correctness of [self], by the induction hypothesis, and
         by transitivity of [visitf_spec], we are done. *)
      eauto using visitf_spec_transitive.
    Qed.

  End Self.

  (* This is the main statement of correctness for the function [visitf_dep]. *)

  Lemma visitf_dep_correct:
    forall s v s' pf,
    visitf_dep s v = exist _ s' pf ->
    visitf_spec s (singleton v) s'.
  Proof.
    (* Expand [s] as a pair [(marks, forest)]. This helps. *)
    intros [ marks forest ]. generalize dependent forest. generalize dependent marks.
    (* The proof is by well-founded induction over the set [marks]. This is the
       same argument that was used to construct [visitf_dep]. *)
    intros ?. induction_wf ih: wf_gt marks.
    introv h.
    (* Unfold one level of the definition of [visitf_dep]. *)
    revert h. unfold visitf_dep. rewrite Fix_eq_ext. fold visitf_dep.
    (* The term [@prove_lt v marks] appears in the goal, applied to a variable
       [heq] of type [marked v marks = false]. By generalizing this term, we
       replace it in the goal with an explicit parameter of type [marked v
       marks = false -> ...]. On the following line, this allows the
       destruction of [marked v marks] to succeed; it would otherwise lead to
       an ill-typed term. Although this may seem pretty simple, it took me a
       couple hours to figure this out. *)
    generalize (@prove_lt v marks).
    destruct (marked v marks) as [|] _eqn.

    (* We now have two cases, which correspond the two branches in the code of
       [visitf_dep]. *)

    (* Case: [v] is marked. *)
    intros _ ?. xx.
    eapply visitf_spec_reflexive.
    eapply marked_spec. eassumption.

    (* Case: [v] is not marked. *)
    intros ?.
    (* The goal mentions a call to [successors]. We replace it with an
       equivalent call to [fold] over some list [ws] of the successors
       of [v]. *)
    destruct (foreach_set_spec (successor_spec v)) as [ ws [ ? [ ? hcall ]]].
    rewrite (foreach_list_spec hcall). clear hcall.
    (* The goal now mentions a call to [fold], which we know produces
       a result of the form [(marks1, forest1), pf1]. Let us introduce
       these names. *)
    match goal with |- context[fold_left ?x1 ?x2 ?x3] =>
      destruct (fold_left x1 x2 x3) as [ [ marks1 forest1 ] pf1 ] _eqn
    end.
    (* In front of us is an equality hypothesis, which we can clarify. *)
    intro. xx.
    (* By applying a suitable lemma, we reduce the goal to an assertion
       about the children [ws] of [v]. *)
    eapply visitf_spec_unmarked; [ eassumption | eassumption | eassumption | ].
    (* To prove this assertion, we apply [fold_self_correct]. *)
    eapply fold_self_correct
      with (P := lift le (mark v marks, Empty V)); [ | eassumption ].
    (* The last sub-goal requires us to prove that [visitf_dep], wrapped
       as a function of a non-dependent type, satisfies its specification.
       Modulo the boilerplate, this is a direct consequence of the induction
       hypothesis. *)
    simpl. intros s2 w. intros.
    destruct (visitf_dep s2 w) as [ s3 ? ] _eqn.
    state. xx.
    eauto using strict_subset_transitive_1, prove_lt.
  Qed.

(* ---------------------------------------------------------------------------- *)

  (* Re-packaging. *)

  (* We re-package [visitf_dep] as a function with a simple (non-dependent) type. *)

  Definition visitf (s0 : state) (v : V) : state :=
    proj1_sig (visitf_dep s0 v).

  (* The re-packaged function [visitf] is correct. *)

  Lemma visitf_correct:
    forall s v s',
    visitf s v = s' ->
    visitf_spec s (singleton v) s'.
  Proof.
    unfold visitf. introv h.
    destruct (visitf_dep s v) as [ ? ? ] _eqn. simpl in h. subst.
    eauto using visitf_dep_correct.
  Qed.

  (* It is now easy to obtain a specification for [fold_left visitf]. *)

  Lemma fold_visitf_correct:
    forall s' ws s,
    fold_left visitf ws s = s' ->
    visitf_spec s (lsupport ws) s'.
  Proof.
    (* We are really re-proving [fold_self_correct], where [self] is
       instantiated with [visitf]. It is simpler to re-do the proof
       here than to abstract [fold_self_correct] so as to re-use it. *)
    induction ws as [| w ws ]; simpl; introv h.
    (* Case Nil. *)
    subst. state.
    solve [ eauto using visitf_spec_reflexive with mysets ].
    (* Case Cons. *)
    destruct (visitf s w) as [ ? ? ] _eqn.
    eauto using visitf_spec_transitive, visitf_correct.
  Qed.

(* ---------------------------------------------------------------------------- *)

  (* A special case of [fold_left visitf], where the initial state consists of
     an empty set of marks and an empty forest. We unfold [visitf_spec] and
     specialize it for this case. *)

  (* This is the simplified specification. [rs] is the set of vertices that
     are offered as roots, i.e., the set of starting points for the DFS
     traversal. [omarked] is the final set of marked vertices. [f] is the
     forest that is constructed. *)

  Definition visitf_initial_post (rs omarked : set V) (f : forest V) :=
    (* Every vertex that was offered as a root is finally marked. *)
    subset rs omarked /\
    (* Every root of the forest that was constructed was initially offered as a root. *)
    subset (roots f) rs /\
    (* The forest is valid. *)
    dfs E empty_ omarked (revf f).

  (* This is the code. *)

  Definition fold_visitf_initial (rs : list V) : state :=
    fold_left visitf rs (void, Empty _).

  (* The code is correct. *)

  Lemma fold_visitf_initial_correct:
    forall rs marks' forest',
    fold_visitf_initial rs = (marks', forest') ->
    visitf_initial_post (lsupport rs) (meaning marks') forest'.
  Proof.
    intros.
    forwards f: fold_visitf_correct. eassumption.
    unfold visitf_spec in f. simpl in f. unpack.
    splits.
    eauto.
    eauto with st mysets.
    equates 3. eauto with dfs. symmetry. eapply void_spec.
  Qed.

(* ---------------------------------------------------------------------------- *)

  (* If the list of roots [rs] is explicit, then [fold_visitf_initial], which
     was defined and specified above, can be used. Sometimes, though, it is
     preferable for this list to be implicit, so as to avoid constructing it
     in memory. *)

  (* We assume that the function [foreach root] allows iterating over the
     list [rs]. *)
  
  Variable root : FOREACH V.
  Variable rs : list V.
  Variable root_spec : FOREACH_LIST_SPEC root rs.

  (* This is the code. *)

  Definition fold_visitf_initial_implicit : state :=
    foreach root (void, Empty _) visitf.

  (* The code is correct. *)

  Lemma fold_visitf_initial_implicit_correct:
    forall marks' forest',
    fold_visitf_initial_implicit = (marks', forest') ->
    visitf_initial_post (lsupport rs) (meaning marks') forest'.
  Proof.
    unfold fold_visitf_initial_implicit.
    rewrite (foreach_list_spec root_spec).
    eauto using fold_visitf_initial_correct.
  Qed.

End DFSExec.

(* Optional.
Extraction visitf_dep.
Extraction visitf.
Extraction fold_visitf_initial.
Extraction fold_visitf_initial_implicit.
*)

