(* This file defines a type [FOREACH V] for a higher-order function
   that allows iterating over a collection of elements of type [V].
   It also defines two specifications, [FOREACH_LIST_SPEC] and
   [FOREACH_SET_SPEC], which claim that such a function correctly
   iterates over an underlying list or set of elements. *)

Set Implicit Arguments.
Require Import List.
Require Import MySets.

(* The function [foreach] applies its argument [f] at each element of the
   collection in succession. *)

Record FOREACH (V : Type) := MkFOREACH {

  foreach : forall A, A -> (A -> V -> A) -> A

}.

(* The specification [FOREACH_LIST_SPEC F vs] claims that a call to [foreach F]
     is equivalent to an application of [fold_left] to the list [vs]. *)

Record FOREACH_LIST_SPEC (V : Type) (F : FOREACH V) (vs : list V) := MkFOREACH_LIST_SPEC {

  foreach_list_spec:
    forall A (a : A) f,
    foreach F a f = fold_left f vs a

}.

(* The specification [FOREACH_SET_SPEC F vset] claims that a call to [foreach F]
     is equivalent to an application of [fold_left] to some list [vs] whose set
     of elements is [vset]. *)

Record FOREACH_SET_SPEC (V : Type) (F : FOREACH V) (vset : set V) := MkFOREACH_SET_SPEC {

  foreach_set_spec:
    exists vs,
    subset (lsupport vs) vset /\
    subset vset (lsupport vs) /\
    FOREACH_LIST_SPEC F vs

}.
